class ChartDecorator(object):
    @property
    def width_left(self):
        """
        Required chart margin at left for decorate

        :rtype: int
        """
        return 0

    @property
    def width_top(self):
        """
        Required chart margin at top for decorate

        :rtype: int
        """
        return 0

    @property
    def width_right(self):
        """
        Required chart margin at right for decorate

        :rtype: int
        """
        return 0

    @property
    def width_bottom(self):
        """
        Required chart margin at bottom for decorate

        :rtype: int
        """
        return 0

    def decorate(self, lines):
        """
        Decorate the rendered data

        :param lines: Lines of rendered data
        :type lines: list[str or unicode]
        :return: Decorated lines
        :rtype: list[str or unicode]
        """
        raise NotImplementedError


class AxisDecorator(ChartDecorator):
    BORDER_VERTICAL = u'\U00002502'
    BORDER_HORIZONTAL = u'\U00002500'
    BORDER_CROSSING = u'\U0000253c'

    @property
    def width_left(self):
        return 1

    @property
    def width_bottom(self):
        return 1

    def decorate(self, lines):
        width = len(lines[0])

        decorated_lines = [
            u'{border}{data}'.format(
                border=AxisDecorator.BORDER_VERTICAL,
                data=line)
            for line in lines
        ]

        decorated_lines.append(
            u'{origin}{border}'.format(
                origin=AxisDecorator.BORDER_CROSSING,
                border=AxisDecorator.BORDER_HORIZONTAL * width)
        )

        return decorated_lines


class FrameDecorator(ChartDecorator):
    BORDER_VERTICAL = u'\U00002503'
    BORDER_HORIZONTAL = u'\U00002501'
    BORDER_CORNERS = (u'\U0000250f', u'\U00002513', u'\U00002517', u'\U0000251b')

    @property
    def width_left(self):
        return 1

    @property
    def width_top(self):
        return 1

    @property
    def width_right(self):
        return 1

    @property
    def width_bottom(self):
        return 1

    def decorate(self, lines):
        width = len(lines[0])

        decorated_lines = [
            u'{border}{data}{border}'.format(
                border=self.BORDER_VERTICAL,
                data=line)
            for line in lines
        ]

        decorated_lines.insert(
            0,
            u'{top_left}{border}{top_right}'.format(
                top_left=self.BORDER_CORNERS[0],
                top_right=self.BORDER_CORNERS[1],
                border=self.BORDER_HORIZONTAL * width
            )
        )

        decorated_lines.append(
            u'{bottom_left}{border}{bottom_right}'.format(
                bottom_left=self.BORDER_CORNERS[2],
                bottom_right=self.BORDER_CORNERS[3],
                border=self.BORDER_HORIZONTAL * width
            )
        )

        return decorated_lines


class DoubleFrameDecorator(FrameDecorator):
    BORDER_VERTICAL = u'\U00002551'
    BORDER_HORIZONTAL = u'\U00002550'
    BORDER_CORNERS = (u'\U00002554', u'\U00002557', u'\U0000255a', u'\U0000255d')


class ThinFrameDecorator(FrameDecorator):
    BORDER_VERTICAL = u'\U00002502'
    BORDER_HORIZONTAL = u'\U00002500'
    BORDER_CORNERS = (u'\U0000250c', u'\U00002510', u'\U00002514', u'\U00002518')


class PaperDecorator(ChartDecorator):
    TOP_LEFT = (
        u'          ',
        u'          ',
        u'          ',
        u'          ',
        u'  .-------',
        u' /  .-.   ',
        u'|  /   \\  ',
        u'| |\\_.  | ',
        u'|\\|  | /| ',
        u'| `---\' | ',
    )

    LEFT = u'|       | '

    BOTTOM_LEFT = (
        u'|       |-',
        u'\\       | ',
        u' \\     /  ',
        u'  `---\'   ',
    )

    TOP_BOTTOM = u'-'

    TOP_RIGHT = (
        u'  .---.  ',
        u' /  .  \\ ',
        u'|\\_/|   |',
        u'|   |  /|',
        u'------\' |',
    )

    RIGHT = u'  |'

    BOTTOM_RIGHT = (
        u'  |',
        u' / ',
        u'\'  ',
    )

    START_X = 10
    START_Y = 5
    RIGHT_SHIFT = 6

    @property
    def width_left(self):
        return 10

    @property
    def width_top(self):
        return 5

    @property
    def width_right(self):
        return 4

    @property
    def width_bottom(self):
        return 4

    def decorate(self, lines):
        lines_width = max([len(line) for line in lines])

        lines = [u'{0}{1}{2}'.format(self.LEFT, line.ljust(lines_width), self.RIGHT) for line in lines]
        border = u'{0}{1}{2}'.format(self.LEFT[:-1], self.TOP_BOTTOM * (lines_width + 2), self.RIGHT[1:])
        lines.insert(0, border)
        lines.append(border)

        decorated_lines = []
        self._merge(decorated_lines, 0, self.START_Y - 1, lines)
        self._merge(decorated_lines, 0, 0, self.TOP_LEFT)
        self._merge(decorated_lines, self.START_X + lines_width - self.RIGHT_SHIFT, 0, self.TOP_RIGHT)
        self._merge(decorated_lines, 0, self.START_Y + len(lines) - 2, self.BOTTOM_LEFT)
        self._merge(decorated_lines, self.START_X + lines_width, self.START_Y + len(lines) - 4, self.BOTTOM_RIGHT)

        return decorated_lines

    @classmethod
    def _merge(cls, lines, x, y, updater_lines):
        """
        :type lines: list[str or unicode] or tuple[str or unicode]
        :type x: int
        :type y: int
        :type updater_lines: list[str or unicode] or tuple[str or unicode]
        """
        update_height = len(updater_lines)
        update_width = max([len(line) for line in updater_lines])

        while len(lines) < y + update_height:
            lines.append('')

        for i, line in enumerate(lines):
            lines[i] = line.ljust(x + update_width, ' ')

        for i, update_line in zip(range(y, y + update_height), updater_lines):
            lines[i] = lines[i][:x] + update_line + lines[i][x + len(update_line):]



# -*- coding: UTF-8 -*-


class ChartRenderer(object):
    @property
    def char_resolution_horizontal(self):
        """
        Horizontal data-resolution of one character

        :rtype: int
        """
        return 1

    @property
    def char_resolution_vertical(self):
        """
        Horizontal data-resolution of one character

        :rtype: int
        """
        return 1

    def render(self, width, height, values):
        """
        Render values to chart

        :type width: int
        :type height: int
        :param values: Scaled values for chart capacity
        :type values: list[int or float]
        :return: Rendered lines
        :rtype: list[str or unicode]
        """
        raise NotImplementedError


class BoxRenderer(ChartRenderer):
    BLOCKS = (
        u' ',  # 0
        u'▁',  # 1
        u'▂',  # 2
        u'▃',  # 3
        u'▄',  # 4
        u'▅',  # 5
        u'▆',  # 6
        u'▇',  # 7
        u'█',  # 8
    )

    @property
    def char_resolution_vertical(self):
        """
        :rtype: int
        """
        return len(self.BLOCKS) - 1

    def render(self, width, height, values):
        """
        :type width: int
        :type height: int
        :type values: list[int or float]
        :rtype: list[str or unicode]
        """

        output_buffer = [[self.BLOCKS[0]] * width for _ in range(height)]

        x = 0
        resolution = self.char_resolution_vertical
        for value in values:
            y = 0
            while value > resolution:
                self._block_writer(x, y, resolution, output_buffer)
                value -= resolution
                y += 1
            if value > 0:
                self._block_writer(x, y, int(round(value, 0)), output_buffer)
            x += 1

        return [u''.join(line) for line in reversed(output_buffer)]

    def _block_writer(self, x, y, block_variant, output_buffer):
        """
        :type x: int
        :type y: int
        :type block_variant: int
        :type output_buffer: list[list[str or unicode]]
        """
        output_buffer[y][x] = self.BLOCKS[block_variant]


class SharpRenderer(BoxRenderer):
    BLOCKS = (
        u' ',  # 0
        u'.',  # 1
        u'x',  # 2
        u'#',  # 3
    )


class DrawilleRenderer(ChartRenderer):
    @property
    def char_resolution_horizontal(self):
        return 2

    @property
    def char_resolution_vertical(self):
        return 4

    def render(self, width, height, values):
        """
        :type width: int
        :type height: int
        :type values: list[int or float]
        :rtype: list[str or unicode]
        """
        from drawille import Canvas

        vertical_chart_resolution = height * self.char_resolution_vertical
        horizontal_chart_resolution = width * self.char_resolution_horizontal

        canvas = Canvas()
        x = 0
        for value in values:
            for y in range(1, int(round(value, 0)) + 1):
                canvas.set(x, vertical_chart_resolution - y)
            x += 1

        rows = canvas.rows(min_x=0, min_y=0, max_x=horizontal_chart_resolution, max_y=vertical_chart_resolution)

        if not rows:
            rows = [u'' for _ in range(height)]

        for i, row in enumerate(rows):
            rows[i] = u'{0}'.format(row.ljust(width))

        return rows



# -*- coding: UTF-8 -*-

from __future__ import division
import numpy

from .renderers import BoxRenderer
from .decorators import AxisDecorator

DEFAULT_RENDERER = BoxRenderer()
DEFAULT_DECORATORS = (AxisDecorator(), )


class Tchart(object):
    def __init__(self, width=10, height=5, renderer=DEFAULT_RENDERER, decorators=DEFAULT_DECORATORS):
        """
        :param width: Canvas width
        :type width: int
        :param height: Canvas height
        :type height: int
        :type renderer: tchart.renderers.ChartRenderer
        :type decorators: tuple[tchart.decorators.ChartDecorator] or list[tchart.decorators.ChartDecorator] or None
        """
        self._canvas_width = width
        self._canvas_height = height
        self._chart_renderer = renderer
        self._decorators = decorators or ()

        margin_left = margin_top = margin_right = margin_bottom = 0
        if decorators:
            margin_left = sum([decorator.width_left for decorator in decorators])
            margin_top = sum([decorator.width_top for decorator in decorators])
            margin_right = sum([decorator.width_right for decorator in decorators])
            margin_bottom = sum([decorator.width_bottom for decorator in decorators])

        margin_horizontal = margin_left + margin_right
        margin_vertical = margin_top + margin_bottom

        self._chart_width = self._canvas_width - margin_horizontal
        self._chart_height = self._canvas_height - margin_vertical
        self._chart_resolution_horizontal = self._chart_width * self._chart_renderer.char_resolution_horizontal
        self._chart_resolution_vertical = self._chart_height * self._chart_renderer.char_resolution_vertical

        if self._chart_width < 1 or self._chart_height < 1:
            raise ValueError('Bad dimension; min {min_width}x{min_height}'.format(
                min_width=margin_horizontal + 1,
                min_height=margin_vertical + 1,
            ))

    def render(self, values):
        """
        :type values: list
        :rtype: tuple[str or unicode]
        """
        normalized_values = self._normalize_values(values)

        lines = self._chart_renderer.render(width=self._chart_width, height=self._chart_height,
                                            values=normalized_values)

        for decorator in self._decorators:
            lines = decorator.decorate(lines=lines)

        return lines

    @property
    def canvas_width(self):
        """
        :rtype: int
        """
        return self._canvas_width

    @property
    def canvas_height(self):
        """
        :rtype: int
        """
        return self._canvas_height

    @property
    def chart_width(self):
        """
        :rtype: int
        """
        return self._chart_width

    @property
    def chart_height(self):
        """
        :rtype: int
        """
        return self._chart_height

    def _normalize_values(self, values):
        """
        :type values: list
        :rtype: list
        """
        if not values:
            return []

        minimum = min(values)
        value_range = max(values) - minimum

        vertical_scale = 1
        if value_range > 1:
            vertical_scale = self._chart_resolution_vertical / value_range

        values = DataManipulation.shift_values(values, -minimum)
        values = DataManipulation.horizontal_scale_values(values, self._chart_resolution_horizontal)
        values = DataManipulation.vertical_scale_values(values, vertical_scale)

        return values


class DataManipulation(object):
    @classmethod
    def shift_values(cls, values, shift):
        """
        :type values: list
        :type shift: float
        :rtype: list
        """
        return [value + shift for value in values]

    @classmethod
    def horizontal_scale_values(cls, values, new_width):
        """
        :type values: list
        :type new_width: int
        :rtype: list
        """
        if new_width == 0:
            return []
        if new_width == 1:
            return [numpy.mean(values)]

        return cls._horizontal_scale_values_with_interpolation(values, new_width)

    @classmethod
    def _horizontal_scale_values_with_interpolation(cls, values, new_width):
        """
        :type values: list
        :type new_width: int
        :rtype: list
        """
        old_point_count = len(values)
        if old_point_count == new_width:
            return values

        new_x_partition_count = new_width - 1
        new_x_step_size = (old_point_count - 1) / new_x_partition_count

        new_x_points = [0] * new_width
        if new_x_step_size > 0:
            new_x_points = list(numpy.arange(0, old_point_count - 1, new_x_step_size))
            if len(new_x_points) < new_width:
                new_x_points += [old_point_count - 1]

        new_y_points = numpy.interp(new_x_points, xp=range(old_point_count), fp=values)

        return list(new_y_points)

    @classmethod
    def vertical_scale_values(cls, values, scale):
        """
        :type values: list
        :type scale: float
        :rtype: list
        """
        return [value * scale for value in values]
